library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

package T1_T2_interface_modules is

component t2tot1_serialiser
	port (
		Reset : in std_logic;
		parallel_in : in std_logic_vector(3 downto 0);
		Clk40 : in std_logic;
		Clk160 : in std_logic;
		serial_out : out std_logic);
end component t2tot1_serialiser;

component t2tot1_deserialiser
	port (
		Reset : in std_logic;
		parallel_out : out std_logic_vector(3 downto 0);
		deser_aligned : out std_logic;
		Clk40 : in std_logic;
		Clk160 : in std_logic;
		serial_in : in std_logic);
end component t2tot1_deserialiser;

component t2tot1_interface
	port (
		Reset : in std_logic;
		Clk40 : in std_logic;
		Clk160 : in std_logic;
		ttc_data : in std_logic;
		invert_t3_signals : in std_logic_vector(4 downto 0);
		tcds_mode : in std_logic;										-- TCDS mode if set to 1, else AMC13 mode (default to 0)
		GT_select : in std_logic;										-- TCDS mode if set to 1, else AMC13 mode (default to 0)
		T3_mosi_TTC_orbit_n : in std_logic;							-- port name speaks for itself
		T3_miso_TTC_orbit_p : in std_logic;							-- port name speaks for itself
		T3_TCDS_aux1_GT1_L1A : in std_logic;						-- port name speaks for itself
		T3_TCDS_aux2_GT2_L1A : in std_logic;						-- port name speaks for itself
		T3_cs_nim_trigger : in std_logic;							-- port name speaks for itself
		T3_sck_technical_trigger : out std_logic;					-- port name speaks for itself
		v2s_p_tcds_l1a_amc13_tech_trigger : inout std_logic;	-- carries the active L1A to T1 in TCDS mode and the technical trigger from T1 in AMC13 mode
		v2s_n_tcds_misc : out std_logic);							-- carries the orbit, nim_trigger, GT1_L1A, GT2_L1A to T1
end component t2tot1_interface;

component t1tot2_interface
	port (
		Reset : in std_logic;
		Clk40 : in std_logic;
		Clk160 : in std_logic;
		tcds_mode : in std_logic;										-- TCDS mode if set to 1, else AMC13 mode (default to 0)
		v2s_p_tcds_l1a_amc13_tech_trigger : inout std_logic;	-- carries the active L1A to T1 in TCDS mode and the technical trigger from T1 in AMC13 mode
		v2s_n_tcds_misc : in std_logic;
		orbit : out std_logic;
		GT_L1A : out std_logic;											-- selected (in T2) active GT trigger; shortest latency
		GT1_L1A : out std_logic;										-- GT1 L1A; longer latency to be used for monitoring
		GT2_L1A : out std_logic;										-- GT2 L1A; longer latency to be used for monitoring
		nim_trigger : out std_logic;									-- auxiliary "trigger" from T3 NIM input;
		technical_trigger : in std_logic);							-- carries the orbit, nim_trigger, GT1_L1A, GT2_L1A to T1
end component t1tot2_interface;

end T1_T2_interface_modules;

-------------------------------------------------------------------------------------------------------------
-------------------------------------------------------------------------------------------------------------
-------------------------------------------------------------------------------------------------------------

-- T2 to T1 Output --
-- Encodes orbit, aux_in, gt1_l1a and gt2_l1a serially with 160 MHz clock.
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

entity t2tot1_serialiser is
	port (
		Reset : in std_logic;
		parallel_in : in std_logic_vector(3 downto 0);
		Clk40 : in std_logic;
		Clk160 : in std_logic;
		serial_out : out std_logic);
end entity t2tot1_serialiser;

architecture t2tot1_serialiser_arch of t2tot1_serialiser is

-- INTERNAL SIGNALS --
signal parallel_in_q : std_logic_vector(3 downto 0);
signal parallel_in_q1 : std_logic_vector(3 downto 0);
signal parallel_in_q2 : std_logic_vector(3 downto 0);
signal reset_q40 : std_logic;
signal reset_q160 : std_logic;
signal phase : integer range 0 to 3;

begin

ffd40: process(clk40)
begin									   
	if rising_edge(Clk40) then		
		for i in 0 to 3 loop
			parallel_in_q1(i) <= parallel_in(i);
			parallel_in_q2(i) <= parallel_in_q1(i);
		end loop;
		parallel_in_q(0) <= not (parallel_in_q1(0) and not parallel_in_q2(0)); -- orbit signal limited to one clock cycle
		parallel_in_q(1) <= parallel_in_q1(1) and not parallel_in_q2(1);-- NIM triggers may not span multiple clocks
		parallel_in_q(2) <= parallel_in_q1(2);-- GT triggers MAY span multiple clocks
		parallel_in_q(3) <= parallel_in_q1(3);-- GT triggers MAY span multiple clocks
		reset_q40 <= reset;
	else null;
	end if;
end process ffd40;


ffd160: process(clk160)
begin									   
	if rising_edge(Clk160) then		
		if reset_q160 = '1' then
			phase <= 1;
		elsif phase = 3 then
			phase <= 0;
		else
			phase <= phase + 1;
		end if;
		serial_out <= parallel_in_q(phase);
		reset_q160 <= reset_q40;
  end if;
end process ffd160;

end t2tot1_serialiser_arch;

-- T2 to T1 Input --
-- decodes orbit, aux_in, gt1_l1a and gt2_l1a received serially with 160 MHz clock.
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

entity t2tot1_deserialiser is
	port (
		Reset : in std_logic;
		parallel_out : out std_logic_vector(3 downto 0);
		deser_aligned : out std_logic;
		Clk40 : in std_logic;
		Clk160 : in std_logic;
		serial_in : in std_logic);
end entity t2tot1_deserialiser;

architecture t2tot1_deserialiser_arch of t2tot1_deserialiser is

-- INTERNAL SIGNALS --
signal parallel_out_d : std_logic_vector(3 downto 0);
signal parallel_out_q1 : std_logic_vector(3 downto 0);
signal parallel_out_q2 : std_logic_vector(3 downto 0);
signal reset_q40 : std_logic;
signal reset_q160 : std_logic;
signal alignment_counter : integer range 0 to 63;
signal phase : integer range 0 to 3;
signal change_phase : std_logic;
signal aligned : std_logic;
signal deadtime_counter : integer range 0 to 127;

begin

ffd40: process(clk40)
begin									   
	if rising_edge(Clk40) then		
		aligned <= '0';
		if aligned = '0' then
			parallel_out <= (others => '0');
		else
			parallel_out <= parallel_out_q1;
			parallel_out(0) <= not parallel_out_q1(0);
		end if;
		if alignment_counter < 40 then
			deadtime_counter <= 100;
		elsif deadtime_counter = 0 then
			deadtime_counter <= deadtime_counter;
			aligned <= '1';
		else
			deadtime_counter <= deadtime_counter - 1;
		end if;
		reset_q40 <= reset;
	else null;
	end if;
end process ffd40;

deser_aligned <= aligned;

ffd160: process(clk160)
begin									   
	if rising_edge(Clk160) then		
		if alignment_counter < 10 then
			change_phase <= '1';
		else null;
		end if;
		if reset_q160 = '1' then
			phase <= 1;
		elsif phase = 3 and change_phase = '1' then
			phase <= 1;
			alignment_counter <= 60;
			change_phase <= '0';
		elsif phase = 3 then
			phase <= 0;
		else
			phase <= phase + 1;
		end if;
		parallel_out_d(phase) <= serial_in;
		if phase = 0 then
			parallel_out_q1 <= parallel_out_d;
		else
			parallel_out_q1 <= parallel_out_q1;
		end if;
		if parallel_out_q1(0) = '1' and alignment_counter < 50 then
			alignment_counter <= alignment_counter + 1;
		elsif parallel_out_q1(0) = '0' and alignment_counter > 0 then
			alignment_counter <= alignment_counter - 1;
		else null;
		end if;
		reset_q160 <= reset_q40;
	end if;
end process ffd160;

end t2tot1_deserialiser_arch;


-------------------------------------------------------------------------------------------------------------
-------------------------------------------------------------------------------------------------------------
-------------------------------------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
library work;
use work.T1_T2_interface_modules.ALL;
library unisim;
use unisim.vcomponents.all;

entity t2tot1_interface is
	port (
		Reset : in std_logic;
		parallel_out : out std_logic_vector(3 downto 0);
		Clk40 : in std_logic;
		Clk160 : in std_logic;
		ttc_data : in std_logic;
		invert_t3_signals : in std_logic_vector(4 downto 0);
		tcds_mode : in std_logic;										-- TCDS mode if set to 1, else AMC13 mode
		GT_select : in std_logic;										-- TCDS mode if set to 1, else AMC13 mode
		T3_mosi_TTC_orbit_n : in std_logic;							-- port name speaks for itself
		T3_miso_TTC_orbit_p : in std_logic;							-- port name speaks for itself
		T3_TCDS_aux1_GT1_L1A : in std_logic;						-- port name speaks for itself
		T3_TCDS_aux2_GT2_L1A : in std_logic;						-- port name speaks for itself
		T3_cs_nim_trigger : in std_logic;							-- port name speaks for itself
		T3_sck_technical_trigger : out std_logic;					-- port name speaks for itself
		v2s_p_tcds_l1a_amc13_tech_trigger : inout std_logic;	-- carries the active L1A to T1 in TCDS mode and the technical trigger from T1 in AMC13 mode
		v2s_n_tcds_misc : out std_logic);								-- carries the orbit, nim_trigger, GT1_L1A, GT2_L1A to T1
end entity t2tot1_interface;

architecture t2tot1_interface_arch of t2tot1_interface is

	signal parallel_in : std_logic_vector(3 downto 0);
	signal orbit : std_logic;

begin

i_orbit: IBUFDS
	port map (
		O	=> orbit,
		I	=> T3_miso_TTC_orbit_p,		-- Connect to package pin H2
		IB	=> T3_mosi_TTC_orbit_n);   -- Connect to package pin H1


parallel_in(0) <= not orbit xor invert_t3_signals(0);
parallel_in(1) <= not T3_cs_nim_trigger xor invert_t3_signals(1);
parallel_in(2) <= not T3_TCDS_aux1_GT1_L1A xor invert_t3_signals(2);
parallel_in(3) <= not T3_TCDS_aux2_GT2_L1A xor invert_t3_signals(3);

t2tot1_serialiser_inst : t2tot1_serialiser
	port map(
		Reset => Reset,
		parallel_in => parallel_in,
		Clk40 => clk40,
		Clk160 => Clk160,
		serial_out => v2s_n_tcds_misc);



v2s_p_tcds_l1a_amc13_tech_trigger <=	not T3_TCDS_aux1_GT1_L1A xor invert_t3_signals(4) when tcds_mode = '1' and GT_select = '0' else
													not T3_TCDS_aux2_GT2_L1A xor invert_t3_signals(4) when tcds_mode = '1' and GT_select = '1' else
													'Z';

T3_sck_technical_trigger <=	v2s_p_tcds_l1a_amc13_tech_trigger when tcds_mode = '0' else
										TTC_data; -- TTC data spy when in TCDS mode; this is really L1A and Bgo.
										--orbit;--TTC_data; -- TTC data spy when in TCDS mode; this is really L1A and Bgo.



end t2tot1_interface_arch;

-------------------------------------------------------------------------------------------------------------
-------------------------------------------------------------------------------------------------------------
-------------------------------------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
library work;
use work.T1_T2_interface_modules.ALL;
library unisim;
use unisim.vcomponents.all;

entity t1tot2_interface is
	port (
		Reset : in std_logic;
		Clk40 : in std_logic;
		Clk160 : in std_logic;
		tcds_mode : in std_logic;										-- TCDS mode if set to 1, else AMC13 mode (default to 0)
		v2s_p_tcds_l1a_amc13_tech_trigger : inout std_logic;	-- carries the active L1A to T1 in TCDS mode and the technical trigger from T1 in AMC13 mode
		v2s_n_tcds_misc : in std_logic;
		orbit : out std_logic;
		GT_L1A : out std_logic;											-- selected (in T2) active GT trigger; shortest latency
		GT1_L1A : out std_logic;										-- GT1 L1A; longer latency to be used for monitoring
		GT2_L1A : out std_logic;										-- GT2 L1A; longer latency to be used for monitoring
		nim_trigger : out std_logic;									-- auxiliary "trigger" from T3 NIM input;
		technical_trigger : in std_logic);							-- carries the orbit, nim_trigger, GT1_L1A, GT2_L1A to T1
end entity t1tot2_interface;

architecture t1tot2_interface_arch of t1tot2_interface is

	signal parallel_out : std_logic_vector(3 downto 0);

begin

GT_L1A <= v2s_p_tcds_l1a_amc13_tech_trigger when tcds_mode = '1' else '0';

v2s_p_tcds_l1a_amc13_tech_trigger <= technical_trigger when tcds_mode = '0' else 'Z';

t2tot1_deserialiser_arch : t2tot1_deserialiser
  port map(
		Reset => Reset,
		parallel_out => parallel_out,
		Clk40 => clk40,
		Clk160 => Clk160,
		serial_in => v2s_n_tcds_misc);

orbit <= parallel_out(0);
nim_trigger <= parallel_out(1);
GT1_L1A <= parallel_out(2);
GT2_L1A <= parallel_out(3);

end t1tot2_interface_arch;

-------------------------------------------------------------------------------------------------------------
-------------------------------------------------------------------------------------------------------------
-------------------------------------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use IEEE.NUMERIC_STD.ALL;
library work;
use work.T1_T2_interface_modules.ALL;

entity T1_T2_interface_serdes_tb is
end entity T1_T2_interface_serdes_tb;

architecture T1_T2_interface_serdes_tb_arch of T1_T2_interface_serdes_tb is

	signal all_done : std_logic;

	signal Reset_transmitter : std_logic;
	signal Reset_receiver : std_logic;
	signal Clk40 : std_logic;
	signal Clk160 : std_logic;
	signal parallel_in : std_logic_vector(3 downto 0);
	signal parallel_out : std_logic_vector(3 downto 0);
	signal serial_data : std_logic;

begin


   clocks: process
      begin
			for j in 1 to 100000000 loop
				Clk160 <= '0';
				Clk40 <= '0';
				wait for 3.125 NS;
				Clk160 <= '1';
				Clk40 <= '0';
				wait for 3.125 NS;
				Clk160 <= '0';
				Clk40 <= '0';
				wait for 3.125 NS;
				Clk160 <= '1';
				Clk40 <= '1';
				wait for 3.125 NS;
				Clk160 <= '0';
				Clk40 <= '1';
				wait for 3.125 NS;
				Clk160 <= '1';
				Clk40 <= '1';
				wait for 3.125 NS;
				Clk160 <= '0';
				Clk40 <= '1';
				wait for 3.125 NS;
				Clk160 <= '1';
				Clk40 <= '0';
				wait for 3.125 NS;
				exit when all_done = '1';
			end loop;
		wait;
	end process clocks;

t2tot1_serialiser_inst : t2tot1_serialiser
	port map(
		Reset => Reset_transmitter,
		parallel_in => parallel_in,
		Clk40 => clk40,
		Clk160 => Clk160,
		serial_out => serial_data);

t2tot1_deserialiser_arch : t2tot1_deserialiser
  port map(
		Reset => Reset_receiver,
		parallel_out => parallel_out,
		Clk40 => clk40,
		Clk160 => Clk160,
		serial_in => serial_data);

test: process
begin
	all_done <= '0';
	Reset_transmitter <= '1';
	Reset_receiver <= '1';
	parallel_in <= "0000";
	wait for 500 NS;

	Reset_transmitter <= '0';
	wait for 200 NS;
	Reset_receiver <= '0';

	wait for 10000 NS;

	wait until falling_edge(clk40); parallel_in(0) <= '1'; wait until falling_edge(clk40); parallel_in(0) <= '0';

	wait for 200 NS;

	wait until falling_edge(clk40); parallel_in(3) <= '1'; wait until falling_edge(clk40); parallel_in(3) <= '0';

	wait for 200 NS;

	wait until falling_edge(clk40); parallel_in <= "1111"; wait until falling_edge(clk40); parallel_in <= "0000";

	wait for 200 NS;

	all_done <= '1';

	wait for 100 NS;

wait;

end process test;


end T1_T2_interface_serdes_tb_arch;

-------------------------------------------------------------------------------------------------------------
-------------------------------------------------------------------------------------------------------------
-------------------------------------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use IEEE.NUMERIC_STD.ALL;
library work;
use work.T1_T2_interface_modules.ALL;
library unisim;
use unisim.vcomponents.all;

entity T1_T2_interface_tb is
end entity T1_T2_interface_tb;

architecture T1_T2_interface_tb_arch of T1_T2_interface_tb is

	signal all_done : std_logic := '0';
	signal ttc_data : std_logic := '0';

	signal Reset_transmitter : std_logic := '0';
	signal Reset_receiver : std_logic := '0';
	signal Clk40 : std_logic := '0';
	signal Clk160 : std_logic := '0';
--	signal ttc_data : std_logic;									
	signal tcds_mode : std_logic := '0';									
	signal GT_select : std_logic := '0';									
	signal orbit_in, orbit_in_p, orbit_in_n : std_logic := '0';	
	signal GT1_L1A_in : std_logic := '0';						
	signal GT2_L1A_in : std_logic := '0';						
	signal nim_trigger_in : std_logic := '0';					
	signal nim_trigger_out : std_logic := '0';					
	signal technical_trigger_in : std_logic := '0';			
	signal technical_trigger_out : std_logic := '0';			
	signal clk40_p : std_logic := '0';							
	signal clk40_n : std_logic := '0';							
	signal serial_data : std_logic := '0';
	signal trigger_wire : std_logic := '0';
	signal orbit_out : std_logic := '0';
	signal GT_L1A_out : std_logic := '0';					
	signal GT1_L1A_out : std_logic := '0';					
	signal GT2_L1A_out : std_logic := '0';					

	signal result : std_logic := '0';					

	signal TTCclk : std_logic := '0';					
	signal TTCclk_bar : std_logic := '0';					
	signal S2V : std_logic := '0';					
	signal S2V_n : std_logic := '0';					
	signal S2V_p : std_logic := '0';					

begin


   clocks: process
      begin
			for j in 1 to 100000000 loop
				Clk160 <= '0';
				Clk40 <= '0';
				wait for 3.125 NS;
				Clk160 <= '1';
				Clk40 <= '0';
				wait for 3.125 NS;
				Clk160 <= '0';
				Clk40 <= '0';
				wait for 3.125 NS;
				Clk160 <= '1';
				Clk40 <= '1';
				wait for 3.125 NS;
				Clk160 <= '0';
				Clk40 <= '1';
				wait for 3.125 NS;
				Clk160 <= '1';
				Clk40 <= '1';
				wait for 3.125 NS;
				Clk160 <= '0';
				Clk40 <= '1';
				wait for 3.125 NS;
				Clk160 <= '1';
				Clk40 <= '0';
				wait for 3.125 NS;
				exit when all_done = '1';
			end loop;
		wait;
	end process clocks;

TTCclk <= Clk40;
TTCclk_bar <= not TTCclk;

i_t1_clock_driver : ODDR2   
	generic map(     
		DDR_ALIGNMENT => "NONE", -- Sets output alignment to "NONE", "C0", "C1"      
		INIT => '0', -- Sets initial state of the Q output to '0' or '1'     
		SRTYPE => "ASYNC") -- Specifies "SYNC" or "ASYNC" set/reset  
	port map (     
		Q => S2V, -- 1-bit output data     
		C0 => TTCclk, -- 1-bit clock input     
		C1 => TTCclk_bar, -- 1-bit clock input     
		CE => '1',  -- 1-bit clock enable input     
		D0 => '0',   -- 1-bit data input (associated with C0)     
		D1 => '1',   -- 1-bit data input (associated with C1)    
		R => '0',    -- 1-bit reset input     
		S => '0'     -- 1-bit set input  
		);

i_S2V: OBUFDS 
	generic map(
		IOSTANDARD => "LVDS_25"
		) 
	port map (
		O => S2V_p, 
		OB => S2V_n, 
		I => S2V
		);

t2tot1_interface_inst : t2tot1_interface
	port map(
		Reset => Reset_transmitter,
		Clk40 => clk40,
		Clk160 => Clk160,
		ttc_data => ttc_data,
		invert_t3_signals => "00000",
		tcds_mode => tcds_mode,
		GT_select => GT_select,
		T3_mosi_TTC_orbit_n => orbit_in_n,
		T3_miso_TTC_orbit_p => orbit_in_p,
		T3_TCDS_aux1_GT1_L1A => GT1_L1A_in,
		T3_TCDS_aux2_GT2_L1A => GT2_L1A_in,
		T3_cs_nim_trigger => nim_trigger_in,
		T3_sck_technical_trigger => technical_trigger_out,
		v2s_p_tcds_l1a_amc13_tech_trigger => trigger_wire,
		v2s_n_tcds_misc => serial_data);

t1tot2_interface_inst : t1tot2_interface
  port map(
		Reset => Reset_receiver,
		Clk40 => clk40,
		Clk160 => Clk160,
		tcds_mode => tcds_mode,										
		v2s_p_tcds_l1a_amc13_tech_trigger => trigger_wire,
		orbit => orbit_out,
		GT_L1A => GT_L1A_out,										
		GT1_L1A => GT1_L1A_out,										
		GT2_L1A => GT2_L1A_out,										
		nim_trigger => nim_trigger_out,							
		technical_trigger => technical_trigger_in,			
		v2s_n_tcds_misc => serial_data);

orbit_in_p <= orbit_in;
orbit_in_n <= not orbit_in;

ttc_data <= not clk40;

result <= orbit_out and GT_L1A_out and GT1_L1A_out and GT2_L1A_out and GT1_L1A_out and nim_trigger_out and technical_trigger_out;

test: process
begin
	all_done <= '0';
	Reset_transmitter <= '1';
	Reset_receiver <= '1';
	orbit_in <= '1';
	GT1_L1A_in <= '0';
	GT2_L1A_in <= '0';
	nim_trigger_in <= '0';
	technical_trigger_in <= '0';
	tcds_mode <= '1';
	GT_select <= '0';

	wait for 500 NS;

	Reset_transmitter <= '0';
	wait for 200 NS;
	Reset_receiver <= '0';

	wait for 10000 NS; 

	wait until falling_edge(clk40); orbit_in <= '0'; wait until falling_edge(clk40); orbit_in <= '1';

	wait for 200 NS;

	wait until falling_edge(clk40); GT1_L1A_in <= '1'; wait until falling_edge(clk40); GT1_L1A_in <= '0';

	wait for 200 NS;

	wait until falling_edge(clk40);
	orbit_in <= '0';
	nim_trigger_in <= '1';
	GT1_L1A_in <= '1';
	GT2_L1A_in <= '1';
	wait until falling_edge(clk40);
	orbit_in <= '1';
	nim_trigger_in <= '0';
	GT1_L1A_in <= '0';
	GT2_L1A_in <= '0';

	wait for 200 NS;

	all_done <= '1';

	wait for 100 NS;

wait;

end process test;


end T1_T2_interface_tb_arch;



